import boto3
import os
import json
import argparse
import subprocess
from shutil import copyfile
from blueleader.webpack import WebpackManager
from blueleader.s3 import S3Manager

"""
Frontend deployment script. 
usage: deploy.py [-h] (--production | --dev) config.json

- Generates a webpack config, setting the publicPath appropriately

- Runs webpack on the generated config.

- Uploads local static assets from `local_static_folder` to 
  `destination_bucket`/`destination_folder`, excluding all generated
  javascript bundles except the current one.

- Updates index.html in the given s3 bucket, generating index.html from 
a templated index file given the current bundle name and location.
"""

def main():
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--production", help="Build and deploy to production", action="store_true")
    group.add_argument("--dev", help="Build and deploy to dev servers", action="store_true")
    group.add_argument("--local", help="Build but don't deploy.", action="store_true")
    parser.add_argument('config_file', help="A blue leader JSON config file")
    args = parser.parse_args()

    config = None
    with open(args.config_file, 'r') as fp:
        config = json.load(fp)

    required = ['region',
                'webpack_config_template',
                'local_webpack_config_template',
                'local_public_path',
                'source_directory',
                'prod_destination_bucket',
                'dev_destination_bucket',
                'destination_folder',
                'local_static_folder',
                'aws_profile']

    for k in required:
        if k not in config:
            raise Exception("Required config paramter %s not in config file %s" % (k, args.config_file))
        
    # Determine our destination bucket and public path
    destination_bucket = None
    public_path = "https://s3-%s.amazonaws.com/%s/%s/" %\
                                            (config['region'],
                                             destination_bucket,
                                             config['destination_folder'])
    webpack_template = config['webpack_config_template']

    if args.production is True:
        print("Deploying for production...")
        config['deploy_mode'] = "production"        
        destination_bucket = config['prod_destination_bucket']
        if "prod_public_path" in config:
            public_path = config["prod_public_path"]
    elif args.dev is True:
        print("Deploying for development...")
        config['deploy_mode'] = "dev"        
        destination_bucket = config['dev_destination_bucket']
        if "dev_public_path" in config:
            public_path = config["dev_public_path"]
    elif args.local is True:
        print("Building locally")
        config['deploy_mode'] = "local"
        public_path = config['local_public_path']            
        webpack_template = config['local_webpack_config_template']            
    else:
        raise Exception("Neither production, dev, nor local flags set")

    print("All assets will reference the static URI: %s" % public_path)

    # Store config in a javascript file accessible to our project
    js_module_file = os.path.join(config['source_directory'], "blueleader.js")
    with open(js_module_file, 'w') as f:
        x = json.dumps(config)
        f.write("export default %s" %x)
        print("Generated javascript module at %s" % js_module_file)

    if args.local is True:
        print("Javascript blueleader config module generated. Exiting.")
        exit(1)
    
    webpack = WebpackManager(webpack_template,
                             public_path = public_path)

    # Generate webpack config, run webpack and gather stats
    print("================ Running Webpack =================")
    res = webpack.run_webpack()
    print(res)
    print("==================================================")
    current_bundle = webpack.current_bundle()
    print("Current webpack bundle: %s" % current_bundle)
    # Upload generated and static assets to S3
    index_output_file = "index.html.out"
    s3 = S3Manager(config['region'], destination_bucket,
                   config['destination_folder'],
                   config['local_static_folder'],
                   config['aws_profile'],
                   public_path
    )
    s3.upload_static()

        

    s3.update_static_routes("index.html.template", "index.html.out",
                            {"bundle_url": current_bundle})

    if args.local is True:
        print("Local configuration complete. Exiting")
        shutil.copyfile("index.html.out", "index.html")
        exit(1)

    s3.upload_file(index_output_file, "index.html",
                   ExtraArgs={'ContentType': "text/html",
                              'CacheControl': "max-age=60"
                   })
    subprocess.Popen("rm %s" % index_output_file,
                            shell=True,
                            stdout=subprocess.PIPE).stdout.read().decode('utf-8')
    
    
if __name__ == '__main__':
    main()



import os
import boto3

class S3Manager(object):
    def __init__(self,
                 aws_region,
                 destination_bucket,
                 destination_folder,
                 local_static_folder,
                 aws_profile,
                 url_override = None
    ):

        self._aws_region = aws_region
        self._destination_bucket = destination_bucket
        self._destination_folder = destination_folder
        self._local_static_folder = local_static_folder
        self._url_override = url_override

        try:
            session = boto3.Session(profile_name=aws_profile,
                                    region_name=aws_region)
        except:
            session = boto3.Session(region_name=aws_region)
        self._s3_client = session.client('s3')
        
    def upload_static(self, ignore=[]):
        """
        Upload local static assets from `local_static_folder` to 
        `destination_bucket`/`destination_folder`
        """
        # Iterate through and upload files
        for filename in os.listdir(self._local_static_folder):
            filepath = self._local_static_folder + filename
            if filename in ignore:
                print("Skipping file %s" % filename)
                continue
            else:
                print("Uploading static asset %s" % filename)
            key = self._destination_folder + "/" + filename
            self.upload_file(filepath, key)

    def public_url(self):
        """
        Returns the public URL of the bucket contents. Overrides with
        the url_override parameter if set (for CDN etc)
        """
        if self._url_override is None:
            return "https://s3-%s.amazonaws.com/%s/%s/%s/" %\
                             (self._aws_region,
                              self._destination_bucket,
                              self._destination_folder)
        else:
            return self._url_override

    def update_static_routes(self, template_file, output_file, mapping):
        """
        Update all static routes in a given file, using mapping.

        Example:
        if mapping = {"bundle_url": "main-ec1246.js"} 
        then we'll transform {bundle_url} into 
          https://s3-us-west-1.amazonaws.com/bucket/folder/main-ec1246.js
          or url_override + "main-ec1256.js"
        """
        # Generate and new file referencing asset locations
        with open(template_file, 'r') as fp:
            template = fp.read()
            rendered = template
            for item in mapping:
                item_url = "%s%s" % (self.public_url(), mapping[item])
                rendered = template.replace("{%s}" % item, item_url)
            with open(output_file, 'w') as fp:
                fp.write(rendered)

    def upload_file(self, local_file, key, ExtraArgs=None):
        self._s3_client.upload_file(local_file,
                                    self._destination_bucket,
                                    key,
                                    ExtraArgs=ExtraArgs
        )
        self._s3_client.put_object_acl(ACL='public-read',
                                       Bucket=self._destination_bucket,
                                       Key=key)



import os
import json
import subprocess

class WebpackManager(object):
    def __init__(self,
                 webpack_config_template,
                 public_path,
                 stats_file = "webpack-stats.json"):
        self._webpack_config_template = webpack_config_template
        self._public_path = public_path
        self._outfile = "generated-webpack-config.out"
        self._stats_file = stats_file

    def current_bundle(self):
        """
        Returns the name of the current javascript bundle
        """
        stats = self._webpack_stats()
        return stats['chunks']['main'][0]['name']

    def _generate_webpack_config(self):
        """
        Generate a webpack config from a template, replacing variables as necessary
        - {public_path}: Replaced with the given public path for static assests (CDN etc)
        """
        config = ""
        with open(self._webpack_config_template, 'r') as fp:
            config = fp.read()
            
        replace = {"public_path": self._public_path}
        for k in replace:
            config = config.replace("{%s}" % k, replace[k])
        
        with open(self._outfile, 'w') as fp:
            return fp.write(config)

    def run_webpack(self):
        """
        Run webpack using the given config file
        """
        self._generate_webpack_config()
        res = subprocess.Popen("webpack --config %s" % self._outfile,
                                shell=True,
                                stdout=subprocess.PIPE).stdout.read().decode('utf-8')
        subprocess.Popen("rm %s" % self._outfile,
                                shell=True,
                                stdout=subprocess.PIPE).stdout.read().decode('utf-8')        
        return res

    def _webpack_stats(self):
        """
        Gather information about most recent webpack run
        """
        with open(self._stats_file, 'r') as fp:
            webpack_stats = json.load(fp)
        if webpack_stats['status'] != 'done':
            raise Exception('Webpack did not successfully complete')
        return webpack_stats
        
